/*
 *************************************************************************
 *
 * serdisp_colour.c
 * colour and colour space specific routines
 *
 *************************************************************************
 *
 * copyright (C) 2003-2006  wolfgang astleitner
 * email     mrwastl@users.sourceforge.net
 *
 *************************************************************************
 * This program is free software; you can redistribute it and/or modify   
 * it under the terms of the GNU General Public License as published by   
 * the Free Software Foundation; either version 2 of the License, or (at  
 * your option) any later version.                                        
 *                                                                        
 * This program is distributed in the hope that it will be useful, but    
 * WITHOUT ANY WARRANTY; without even the implied warranty of             
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      
 * General Public License for more details.                               
 *                                                                        
 * You should have received a copy of the GNU General Public License      
 * along with this program; if not, write to the Free Software            
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA              
 * 02111-1307, USA.  Or, point your browser to                            
 * http://www.gnu.org/copyleft/gpl.html                                   
 *************************************************************************
 */

#include "serdisplib/serdisp_control.h"
#include "serdisplib/serdisp_colour.h"

/* *********************************
   void serdisp_setcolour (dd, x, y, colour)
   *********************************
   set a colour  at position x/y
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   colour ... colour  (format: 0xAARRGGBB)
   *********************************
*/
void serdisp_setcolour(serdisp_t* dd, int x, int y, long colour) {
  serdisp_setpixel(dd, x, y, serdisp_transcolour(dd, colour));
}


/* *********************************
   void serdisp_setgrey (dd, x, y, grey)
   *********************************
   set a grey value at position x/y
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   grey   ... grey value (format: [0, 255])
   *********************************
*/
void serdisp_setgrey(serdisp_t* dd, int x, int y, byte grey) {
  serdisp_setpixel(dd, x, y, serdisp_transgrey(dd, grey));
}


/* *********************************
   long serdisp_getcolour (dd, x, y)
   *********************************
   get colour value at position x / y
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   *********************************
   returns colour value (format: 0xAARRGGBB)
*/
long serdisp_getcolour(serdisp_t* dd, int x, int y) {
  return serdisp_lookupcolour(dd, serdisp_getpixel(dd, x, y));
}


/* *********************************
   byte serdisp_getgrey (dd, x, y)
   *********************************
   get grey value at position x / y
   *********************************
   dd     ... display descriptor
   x      ... x-position
   y      ... y-position
   *********************************
   returns gray value (format: [0, 255])
*/
byte serdisp_getgrey(serdisp_t* dd, int x, int y) {
  return serdisp_lookupgrey(dd, serdisp_getpixel(dd, x, y));
}




/* *********************************
   long serdisp_transcolour(dd, colour)
   *********************************
   translates an ARGB colour value to a hardware dependend value that is suitable for serdisp_setpixel()
   *********************************
   dd     ... display descriptor
   colour ... ARGB colour value
   *********************************
   returns a translated, serdisp_setpixel()-compliant colour value
*/
long serdisp_transcolour(serdisp_t* dd, long colour) {
  return dd->fp_transcolour(dd, colour);
}



/* *********************************
   long serdisp_transgrey(dd, greyvalue)
   *********************************
   translates a grey value to a hardware dependend value that is suitable for serdisp_setpixel()
   *********************************
   dd        ... display descriptor
   greyvalue ... grey value
   *********************************
   returns a translated, serdisp_setpixel()-compliant colour value
*/
long serdisp_transgrey(serdisp_t* dd, byte greyvalue) {
  return dd->fp_transgrey(dd, greyvalue);
}



/* *********************************
   long serdisp_lookupcolour(dd, colour)
   *********************************
   looks up the corresponding ARGB colour value to a serdisp_setpixel()-compliant colour value
   *********************************
   dd     ... display descriptor
   colour ... serdisp_setpixel()-compliant colour value
   *********************************
   returns a translated colour value (format: 0xAARRGGBB)
*/
long serdisp_lookupcolour(serdisp_t* dd, long colour) {
  return dd->fp_lookupcolour(dd, colour);
}



/* *********************************
   byte serdisp_lookupgrey(dd, colour)
   *********************************
   looks up the corresponding grey value to a serdisp_setpixel()-compliant colour value
   *********************************
   dd     ... display descriptor
   colour ... serdisp_setpixel()-compliant colour value
   *********************************
   returns an 8-bit grey value
*/
byte serdisp_lookupgrey(serdisp_t* dd, long colour) {
  return dd->fp_lookupgrey(dd, colour);
}





/* *********************************
   void serdisp_setcoltabentry (dd, idx, colour)
   *********************************
   set a colour in the colour table
   *********************************
   dd     ... display descriptor
   idx    ... index in colour table
   colour ... colour (format: 0xAARRGGBB)
   *********************************
*/
void serdisp_setcoltabentry(serdisp_t* dd, int idx, long colour) {
  if ( (!dd->ctable) || idx >= serdisp_getcolours(dd))
    return;
    
  dd->ctable[idx] = colour;
}


/* *********************************
   long serdisp_getcoltabentry (dd, idx)
   *********************************
   get  a colour from the colour table
   *********************************
   dd     ... display descriptor
   idx    ... index in colour table
   *********************************
   returns colour at index idx
*/
long serdisp_getcoltabentry(serdisp_t* dd, int idx) {
  if ( (!dd->ctable) || idx >= serdisp_getcolours(dd)) 
    return 0L;
  
  return dd->ctable[idx];
}


/* internal use */

/* source: http://www.compuphase.com/cmetric.htm */
long serdisp_colourdistance(long col1, long col2) {
  long r1, g1, b1;
  long r2, g2, b2;
  long distr, distg, distb, meanr;
  
  b1 = col1 & 0xFF;  g1 = (col1 & 0xFF00) >> 8; r1 = (col1 & 0xFF0000) >> 16;
  b2 = col2 & 0xFF;  g2 = (col2 & 0xFF00) >> 8; r2 = (col2 & 0xFF0000) >> 16;
  
  distr = r1-r2;  distg = g1-g2; distb = b1-b2;
  meanr = (r1 + r2) >> 1;
  
  /* return distr*distr + distg*distg + distb*distb; */
  return (((512+meanr)*distr*distr)>>8) + 4*distg*distg + (((767-meanr)*distb*distb)>>8);
}


/* 
 * colour scheme functions
 */


/* *********************************
   long serdisp_transcolour_* (dd, colour)
   *********************************
   translates an ARGB colour to a value suitable for setpixel
   *********************************
   dd     ... display descriptor
   colour ... ARGB colour value
   *********************************
   returns a translated, setpixel-compliant colour value
*/
long serdisp_transcolour_bw (serdisp_t* dd, long colour) {
  /* depth 1 (b/w) anomaly: index 0 == white, index 1 == black */
  return ( (serdisp_ARGB2GREY(colour) > 127 ) ? 0x00000000L : 0x00000001L);
}

long serdisp_transcolour_grey2_4(serdisp_t* dd, long colour) {
  long minidx = 0;
  long mincross = serdisp_colourdistance(colour, serdisp_getcoltabentry(dd, 0));
  long tempcolcross;
  int i;
  
  for (i = 1; i < (1 << dd->depth); i++) {
    tempcolcross = serdisp_colourdistance(colour, serdisp_getcoltabentry(dd, i));
    if (tempcolcross < mincross) {
      mincross = tempcolcross;
      minidx = i;
    }
  }
  return minidx;
}

long serdisp_transcolour_grey8 (serdisp_t* dd, long colour) {
  return ((long)(serdisp_ARGB2GREY(colour)));
}


/* 8 bit colour parts -> 3/3/2 bit by throwing away least significant bits */
/* 0xAARRGGBB -> byte(RRRGGGBB) */
long serdisp_transcolour_rgb332 (serdisp_t* dd, long colour) {
  return  ( (( colour & 0x00E00000L) >> 16) | ((colour & 0x0000E000L) >> 11) | ((colour & 0x000000C0L) >> 6) );
}

/* 8 bit colour part -> 4 bit by throwing away 4 least significant bits */
/* 0xAARRGGBB -> 0x00000RGB */
long serdisp_transcolour_rgb444 (serdisp_t* dd, long colour) {
  return  ( (( colour & 0x00F00000L) >> 12) | ((colour & 0x0000F000L) >> 8) | ((colour & 0x000000F0L) >> 4) );
}



/* *********************************
   long serdisp_transgrey_* (dd, greyvalue)
   *********************************
   translates a grey value to a value suitable for setpixel
   *********************************
   dd        ... display descriptor
   greyvalue ... grey value
   *********************************
   returns a translated, setpixel-compliant colour value
*/
long serdisp_transgrey_bw (serdisp_t* dd, byte greyvalue) {
  /* depth 1 (b/w) anomaly: index 0 == white, index 1 == black */
  return ( ( greyvalue > 127 ) ? 0x00000000L : 0x00000001L);
}

long serdisp_transgrey_grey2_4(serdisp_t* dd, byte greyvalue) {
  long minidx = 0;
  long mincross = serdisp_colourdistance(serdisp_GREY2ARGB(greyvalue), serdisp_getcoltabentry(dd, 0));
  long tempcolcross;
  int i;
  
  for (i = 1; i < (1 << dd->depth); i++) {
    tempcolcross = serdisp_colourdistance(serdisp_GREY2ARGB(greyvalue), serdisp_getcoltabentry(dd, i));
    if (tempcolcross < mincross) {
      mincross = tempcolcross;
      minidx = i;
    }
  }
  return minidx;
}


long serdisp_transgrey_grey8 (serdisp_t* dd, byte greyvalue) {
  return ( (long)greyvalue );
}

long serdisp_transgrey_rgb332 (serdisp_t* dd, byte greyvalue) {
  return ( ((0xE0 & (long)greyvalue)) | ((0xE0 & (long)greyvalue) >> 3) | ((0xC0 & (long)greyvalue) >> 6) );
}

long serdisp_transgrey_rgb444 (serdisp_t* dd, byte greyvalue) {
  return ( ((0xF0 & (long)greyvalue) << 4) | ((0xF0 & (long)greyvalue) ) | ((0xF0 & (long)greyvalue) >> 4) );
}



/* *********************************
   long serdisp_lookupcolour_* (dd, colour)
   *********************************
   looks up the corresponding ARGB colour value to a setpixel-compliant colour value
   *********************************
   dd     ... display descriptor
   colour ... setpixel-compliant colour value
   *********************************
   returns a translated colour value (format: 0xAARRGGBB)
*/
long serdisp_lookupcolour_bw(serdisp_t* dd, long colour) {
  /* depth 1 (b/w) anomaly: index 0 == white, index 1 == black */
  return ( (colour > 0x00000000L) ? 0xFF000000L : 0xFFFFFFFFL);
}

long serdisp_lookupcolour_grey2_4(serdisp_t* dd, long colour) {
  return ( serdisp_getcoltabentry(dd, (int)colour) );
}

long serdisp_lookupcolour_grey8(serdisp_t* dd, long colour) {
  return ( serdisp_GREY2ARGB( (byte)(0x000000FFL | colour) ) );
}


/* byte(RRRGGGBB) -> 0xAARRGGBB */
long serdisp_lookupcolour_rgb332(serdisp_t* dd, long colour) {
  return ( 0xFF1F1F3FL |  ((0xE0L & colour) << 16 ) | ((0x1CL & colour) << 11 ) | ((0x00000003L & colour) << 6 ) );
}


/* 0x00000RGB -> 0xAARRGGBB */
long serdisp_lookupcolour_rgb444(serdisp_t* dd, long colour) {
  return ( 0xFF0F0F0FL |  ((0x00000F00L & colour) << 12 ) | ((0x000000F0L & colour) << 8 ) | ((0x0000000FL & colour) << 4 ) );
}



/* *********************************
   byte serdisp_lookupgrey_* (dd, colour)
   *********************************
   translates a colour/grey value so that it is useable for setpixel
   *********************************
   dd     ... display descriptor
   colour ... setpixel-compliant colour value
   *********************************
   returns an 8-bit grey value
*/
byte serdisp_lookupgrey_bw(serdisp_t* dd, long colour) {
  /* depth 1 (b/w) anomaly: index 0 == white, index 1 == black */
  return ( (colour > 0x00000000L) ? 0 : 255);
}


byte serdisp_lookupgrey_grey2_4(serdisp_t* dd, long colour) {
  return ( serdisp_getcoltabentry(dd, (int)colour) );
}


byte serdisp_lookupgrey_grey8(serdisp_t* dd, long colour) {
  return ( (byte)(0x000000FFL | colour) );
}


byte serdisp_lookupgrey_rgb332(serdisp_t* dd, long colour) {
  return ( serdisp_ARGB2GREY(serdisp_lookupcolour_rgb332(dd, colour)) );
}

byte serdisp_lookupgrey_rgb444(serdisp_t* dd, long colour) {
  return ( serdisp_ARGB2GREY(serdisp_lookupcolour_rgb444(dd, colour)) );
}




